from typing import Optional, List
import numpy as np

# IMPORT ROS
import rospy
from std_msgs.msg import Float32MultiArray, UInt64, Int64, Bool

# IMPORT EAGERX
import eagerx.core.register as register
from eagerx.utils.utils import Msg
from eagerx.core.entities import ResetNode, SpaceConverter
from eagerx.core.constants import process as p

status_map = {"Success": 1, "Timeout": 2, "Collision": 3}


class ResetArm(ResetNode):
    @staticmethod
    @register.spec("ResetArm", ResetNode)
    def spec(
        spec,
        name: str,
        rate: float,
        upper: List[float],
        lower: List[float],
        gripper: bool = True,
        threshold: float = 0.02,
        timeout: float = 4.0,
        process: Optional[int] = p.NEW_PROCESS,
        color: Optional[str] = "grey",
    ):
        """Resets joints & Gripper to goal positions.

        :param spec: Not provided by user.
        :param name: Node name
        :param rate: Rate at which callback is called.
        :param upper: Upper joint limits
        :param lower: Lower joint limits
        :param gripper: Include a gripper
        :param threshold: Closeness to the goal joints before considering the reset to be finished.
        :param timeout: Seconds before considering the reset to be finished, regardless of the closeness. A value of `0` means
                        indefinite.
        :param process: {0: NEW_PROCESS, 1: ENVIRONMENT, 2: ENGINE, 3: EXTERNAL}
        :param color: console color of logged messages. {'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'grey'}
        :return:
        """
        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = ["joints", "in_collision"]
        spec.config.targets = ["goal"]
        spec.config.outputs = ["joints"]
        # Add gripper of also controlled
        if gripper:
            spec.config.outputs.append("gripper")
        # Add custom params
        spec.config.threshold = threshold
        spec.config.timeout = timeout

        # Add converter & space_converter
        spec.inputs.joints.space_converter = SpaceConverter.make("Space_Float32MultiArray", lower, upper, dtype="float32")
        spec.outputs.joints.space_converter = SpaceConverter.make("Space_Float32MultiArray", lower, upper, dtype="float32")
        spec.outputs.gripper.space_converter = SpaceConverter.make("Space_Float32MultiArray", [0], [1], dtype="float32")

    def initialize(self, threshold, timeout):
        self.threshold = threshold
        self.timeout = timeout

        # Create a status publisher (not an output, because asynchronous)
        self.status_pub = rospy.Publisher(f"{self.ns_name}/outputs/status", Int64, queue_size=0)

    @register.states()
    def reset(self):
        self.start = None

    @register.inputs(joints=Float32MultiArray, in_collision=UInt64)
    @register.targets(goal=Float32MultiArray)
    @register.outputs(joints=Float32MultiArray, gripper=Float32MultiArray)
    def callback(self, t_n: float, goal: Msg = None, joints: Msg = None, in_collision: Msg = None):
        if self.start is None:
            self.start = t_n

        # todo: What to do if stuck in collision?
        if len(in_collision.msgs) > 0:
            in_collision = in_collision.msgs[-1].data
        else:
            in_collision = False

        # Process goal & current joint msgs
        joints = np.array(joints.msgs[-1].data, dtype="float32")
        goal = np.array(goal.msgs[-1].data, dtype="float32")

        # Determine done flag
        if np.isclose(joints, goal, atol=self.threshold).all():
            is_done = True
            self.status_pub.publish(Int64(data=status_map["Success"]))
        else:
            if self.timeout > 0 and self.timeout < (t_n - self.start):
                is_done = True
                if in_collision:
                    self.status_pub.publish(Int64(data=status_map["Collision"]))
                else:
                    self.status_pub.publish(Int64(data=status_map["Timeout"]))
            else:
                is_done = False

        # Create output message
        output_msgs = dict(joints=Float32MultiArray(data=goal), gripper=Float32MultiArray(data=[1.0]))
        output_msgs["goal/done"] = Bool(data=is_done)
        return output_msgs



# ROS IMPORTS
from std_msgs.msg import Float32MultiArray

# RX IMPORTS
from eagerx.core import register as register
from eagerx.core.entities import Processor
from eagerx.core.specs import ConverterSpec

# OTHER
import numpy as np


class MirrorAction(Processor):
    MSG_TYPE = Float32MultiArray

    @staticmethod
    @register.spec("MirrorAction", Processor)
    def spec(spec: ConverterSpec, index=0, constant=0, offset=0, scale=1):
        # Initialize spec with default arguments
        params = dict(offset=offset, index=index, scale=scale, constant=constant)
        spec.config.update(params)

    def initialize(self, offset, index, scale, constant):
        self.offset = offset
        self.index = index
        self.scale = scale
        self.constant = constant

    def convert(self, msg):
        action = self.scale * msg.data[self.index] + self.constant
        mirrored = np.array([action, -action], dtype="float32")
        mirrored += self.offset
        return Float32MultiArray(data=mirrored)



import os
import eagerx_interbotix
import rospy
import rosparam
import yaml
from eagerx.utils.node_utils import launch_node
from eagerx.utils.utils import get_param_with_blocking


def get_configs(robot_model: str, motor_config: str, mode_config: str):
    module_path = os.path.dirname(eagerx_interbotix.__file__) + "/../assets/"
    config_path = module_path + "config/"
    try:
        mode_config = mode_config if isinstance(mode_config, str) else f"{config_path}/modes.yaml"
        with open(mode_config, "r") as yamlfile:
            mode_config = yaml.safe_load(yamlfile)
    except IOError:
        rospy.logerr(f"Mode Config File was not found in: {config_path}")
        raise

    try:
        motor_config = motor_config if isinstance(motor_config, str) else f"{config_path}/{robot_model}.yaml"
        with open(motor_config, "r") as yamlfile:
            motor_config = yaml.safe_load(yamlfile)
    except IOError:
        rospy.logerr(f"Motor Config File was not found in: {config_path}")
        raise
    return motor_config, mode_config


def generate_urdf(
    robot_model: str,
    ns="",
    base_link_frame="base_link",
    show_ar_tag=False,
    show_gripper_bar=True,
    show_gripper_fingers=True,
    use_world_frame=True,
    external_urdf_loc="",
    load_gazebo_configs=False,
):
    module_path = os.path.dirname(eagerx_interbotix.__file__) + "/../assets/"
    launch_file = module_path + "xsarm_description.launch"
    cli_args = [
        f"module_path:={module_path}",
        f"robot_model:={robot_model}",
        f"ns:={ns}",
        f"base_link_frame:={base_link_frame}",
        f"show_ar_tag:={show_ar_tag}",
        f"show_gripper_bar:={show_gripper_bar}",
        f"show_gripper_fingers:={show_gripper_fingers}",
        f"use_world_frame:={use_world_frame}",
        f"external_urdf_loc:={external_urdf_loc}",
        f"load_gazebo_configs:={load_gazebo_configs}",
    ]
    launch = launch_node(launch_file, cli_args)
    launch.start()

    # Replace mesh urls
    urdf_key = f"{ns}/{robot_model}/robot_description"
    urdf: str = get_param_with_blocking(urdf_key)
    urdf_sbtd = urdf.replace("package://interbotix_xsarm_descriptions/", module_path)
    rosparam.upload_params(urdf_key, urdf_sbtd)

    return urdf_key

    # Write to tmp file
    #



import pybullet as pyb
import pybullet_data


def cube_and_ground(client_id):
    pyb.setAdditionalSearchPath(pybullet_data.getDataPath(), physicsClientId=client_id)

    # ground plane
    ground_id = pyb.loadURDF(
        "plane.urdf",
        [0, 0, 0],
        useFixedBase=True,
        physicsClientId=client_id,
    )

    # some cubes for obstacles
    cube1_id = pyb.loadURDF("cube.urdf", [-0.35, 0, 0.15], useFixedBase=True, physicsClientId=client_id, globalScaling=0.25)

    # store body indices in a dict with more convenient key names
    bodies = {
        "ground": ground_id,
        "cube1": cube1_id,
    }
    return bodies


def exclude_behind_left_workspace(client_id):
    pyb.setAdditionalSearchPath(pybullet_data.getDataPath(), physicsClientId=client_id)

    # ground plane
    ground_id = pyb.loadURDF(
        "plane.urdf",
        [0, 0, 0],
        useFixedBase=True,
        physicsClientId=client_id,
    )

    # some cubes for obstacles
    cube1_id = pyb.loadURDF("cube.urdf", [-0.75, 0, 0], useFixedBase=True, physicsClientId=client_id, globalScaling=1.0)

    # some cubes for obstacles
    cube2_id = pyb.loadURDF("cube.urdf", [0, 0.75, 0], useFixedBase=True, physicsClientId=client_id, globalScaling=1.0)

    # store body indices in a dict with more convenient key names
    bodies = {
        "ground": ground_id,
        "cube1": cube1_id,
        "cube2": cube2_id,
    }
    return bodies


def exclude_ground(client_id):
    pyb.setAdditionalSearchPath(pybullet_data.getDataPath(), physicsClientId=client_id)

    # ground plane
    ground_id = pyb.loadURDF(
        "plane.urdf",
        [0, 0, 0],
        useFixedBase=True,
        physicsClientId=client_id,
    )

    # store body indices in a dict with more convenient key names
    bodies = {
        "ground": ground_id,
    }
    return bodies


def exclude_ground_minus_2m(client_id):
    pyb.setAdditionalSearchPath(pybullet_data.getDataPath(), physicsClientId=client_id)

    # ground plane
    ground_id = pyb.loadURDF(
        "plane.urdf",
        [0, 0, -2],
        useFixedBase=True,
        physicsClientId=client_id,
    )

    # store body indices in a dict with more convenient key names
    bodies = {
        "ground": ground_id,
    }
    return bodies



from typing import Optional, List, Dict

from collections import deque
from urdf_parser_py.urdf import URDF
import numpy as np

# IMPORT ROS
from std_msgs.msg import Float32MultiArray, UInt64

# IMPORT EAGERX
import eagerx.core.register as register
from eagerx.utils.utils import Msg, get_attribute_from_module
from eagerx.core.entities import Node, SpaceConverter
from eagerx.core.constants import process as p
from eagerx_interbotix.safety import collision as col


class SafePositionControl(Node):
    @staticmethod
    @register.spec("SafePositionControl", Node)
    def spec(
        spec,
        name: str,
        rate: float,
        joints: List[int],
        upper: List[float],
        lower: List[float],
        vel_limit: List[float],
        duration: Optional[float] = None,
        checks: int = 2,
        collision: Dict = None,
        process: Optional[int] = p.NEW_PROCESS,
        color: Optional[str] = "grey",
    ):
        """
        Filters goal joint positions that cause self-collisions or are below a certain height.
        Also check velocity limits.

        :param spec: Not provided by user.
        :param name: Node name
        :param rate: Rate at which callback is called.
        :param joints: joint names
        :param upper: upper joint limits
        :param lower: lower joint limits
        :param vel_limit: absolute velocity joint limits
        :param duration: time (seconds) it takes to reach the commanded positions from the current position.
        :param checks: collision checks performed over the duration.
        :param collision: A dict with the robot & workspace specification.
        :param process: {0: NEW_PROCESS, 1: ENVIRONMENT, 2: ENGINE, 3: EXTERNAL}
        :param color: console color of logged messages. {'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'grey'}
        :return:
        """
        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = ["goal", "current"]
        spec.config.outputs = ["filtered", "in_collision"]

        # Modify custom node params
        spec.config.joints = joints
        spec.config.upper = upper
        spec.config.lower = lower
        spec.config.vel_limit = vel_limit
        spec.config.duration = duration if isinstance(duration, float) else 2.0 / rate
        spec.config.checks = checks

        # Collision detector
        spec.config.collision = collision if isinstance(collision, dict) else None

        # Add converter & space_converter
        spec.inputs.goal.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", "$(config lower)", "$(config upper)", dtype="float32"
        )
        spec.inputs.current.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", "$(config lower)", "$(config upper)", dtype="float32"
        )
        spec.outputs.filtered.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", "$(config lower)", "$(config upper)", dtype="float32"
        )

    def initialize(
        self,
        joints: List[str],
        upper: List[float],
        lower: List[float],
        vel_limit: List[float],
        duration: float,
        checks: int,
        collision: dict,
    ):
        self.joints = joints
        self.upper = np.array(upper, dtype="float")
        self.lower = np.array(lower, dtype="float")
        self.vel_limit = np.array(vel_limit, dtype="float")
        self.duration = duration
        self.checks = checks
        self.dt = 1 / self.rate

        # Setup collision detector
        self.collision = collision
        if collision is not None:
            self.collision_check = True
            # Setup physics server for collision checking
            import pybullet

            if collision.get("gui", False):
                self.col_id = pybullet.connect(pybullet.GUI)
            else:
                self.col_id = pybullet.connect(pybullet.DIRECT)
            # Load workspace
            bodies = get_attribute_from_module(collision["workspace"])(self.col_id)
            # Generate robot urdf (if not a path but a text file)
            r = collision["robot"]
            if r["urdf"].endswith(".urdf"):  # Full path specified
                fileName = r["urdf"]
            else:  # First write to /tmp file (else pybullet cannot load urdf)
                import uuid  # Use this to generate a unique filename

                fileName = f"/tmp/{str(uuid.uuid4())}.urdf"
                with open(fileName, "w") as file:
                    file.write(r["urdf"])
            # Load robot
            bodies["robot"] = pybullet.loadURDF(
                fileName,
                basePosition=r.get("basePosition", None),
                baseOrientation=r.get("baseOrientation", None),
                useFixedBase=r.get("useFixedBase", True),
                flags=r.get("flags", 0),
                physicsClientId=self.col_id,
            )
            urdf = URDF.from_xml_file(fileName)
            # Determine collision pairs
            self_collision_pairs, workspace_pairs = col.get_named_collision_pairs(bodies, urdf, joints)
            # Create collision detector
            self.self_collision = col.CollisionDetector(self.col_id, bodies, joints, self_collision_pairs)
            self.workspace = col.CollisionDetector(self.col_id, bodies, joints, workspace_pairs)
            # Set distance at which objects are considered in collision.
            self.margin = collision.get("margin", 0.0)
            # self._test_collision_tester(joints)
        else:
            self.collision_check = False

    def _test_collision_tester(self, joints):
        while True:
            # compute shortest distances for a random configuration
            q = np.pi * (np.random.random(len(joints)) - 0.5)
            in_col = self.self_collision.in_collision(q)

            print(f"[self_collision] In collision = {in_col}")

            in_col = self.workspace.in_collision(q)

            print(f"[workspace] In collision = {in_col}")

            # wait for user to press enter to continue
            input("Press <enter>!")

    @register.states()
    def reset(self):
        self.safe_poses = deque(maxlen=10)
        self.consecutive_unsafe = 0

    @register.inputs(goal=Float32MultiArray, current=Float32MultiArray)
    @register.outputs(filtered=Float32MultiArray, in_collision=UInt64)
    def callback(self, t_n: float, goal: Msg = None, current: Msg = None):
        goal = np.array(goal.msgs[-1].data, dtype="float32")
        current = np.array(current.msgs[-1].data, dtype="float32")

        # Setpoint last safe position
        if self.collision_check:
            if (
                self.self_collision.in_collision(q=current, margin=self.margin)
                and self.self_collision.get_distance().min() < 0
            ):
                # rospy.loginfo(f"[self_collision]: margin = {self.margin} | ds = {self.self_collision.get_distance().min()}")
                in_collision = UInt64(data=1)
            elif self.workspace.in_collision(margin=self.margin) and self.workspace.get_distance().min() < 0:
                # rospy.loginfo(f"[workspace]: margin = {self.margin} | ds = {self.workspace.get_distance().min()}")
                in_collision = UInt64(data=2)
            else:
                in_collision = UInt64(data=0)
                if self.consecutive_unsafe == 0:
                    self.safe_poses.append(current)
        else:
            in_collision = UInt64(data=0)

        # Clip to joint limits
        filtered = np.clip(goal, self.lower, self.upper, dtype="float32")
        # Reduce goal to vel_limit
        diff = filtered - current
        too_fast = np.abs(diff / (self.duration)) > self.vel_limit
        if np.any(too_fast):
            filtered[too_fast] = current[too_fast] + np.sign(diff[too_fast]) * self.dt * self.vel_limit[too_fast]

        if self.collision_check:
            # Linearly interpolate for intermediate joint configurations
            t = np.linspace(self.dt / self.checks, self.dt, self.checks)
            interp = np.empty((current.shape[0], self.checks), dtype="float32")
            diff = filtered - current

            for i in range(current.shape[0]):
                interp[i][:] = np.interp(t, [0, self.duration], [current[i], current[i] + diff[i] * 1.02])

            for i in range(self.checks):
                if self.self_collision.in_collision(q=interp[:, i], margin=self.margin):
                    filtered = self.safe_poses[-1] if len(self.safe_poses) > 0 else current
                    self.consecutive_unsafe += 1
                    break
                elif self.workspace.in_collision(margin=self.margin):
                    filtered = self.safe_poses[-1] if len(self.safe_poses) > 0 else current
                    self.consecutive_unsafe += 1
                    break
                if i + 1 == self.checks:  # If all checks were successful (i.e. we did not break for-loop).
                    self.consecutive_unsafe = 0
        else:
            self.consecutive_unsafe = 0
            in_collision = UInt64(data=0)
        return dict(filtered=Float32MultiArray(data=filtered), in_collision=in_collision)


class SafeVelocityControl(Node):
    @staticmethod
    @register.spec("SafeVelocityControl", Node)
    def spec(
        spec,
        name: str,
        rate: float,
        joints: List[int],
        upper: List[float],
        lower: List[float],
        vel_limit: List[float],
        duration: Optional[float] = None,
        checks: int = 2,
        collision: Dict = None,
        process: Optional[int] = p.NEW_PROCESS,
        color: Optional[str] = "grey",
    ):
        """
        Filters goal joint positions that cause self-collisions or are below a certain height.
        Also check velocity limits.

        :param spec: Not provided by user.
        :param name: Node name
        :param rate: Rate at which callback is called.
        :param joints: joint names
        :param upper: upper joint limits
        :param lower: lower joint limits
        :param vel_limit: absolute velocity joint limits
        :param duration: time (seconds) it takes to reach the commanded positions from the current position.
        :param checks: collision checks performed over the duration.
        :param collision: A dict with the robot & workspace specification.
        :param process: {0: NEW_PROCESS, 1: ENVIRONMENT, 2: ENGINE, 3: EXTERNAL}
        :param color: console color of logged messages. {'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'grey'}
        :return:
        """
        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = ["goal", "position", "velocity"]
        spec.config.outputs = ["filtered", "in_collision"]

        # Modify custom node params
        spec.config.joints = joints
        spec.config.upper = upper
        spec.config.lower = lower
        spec.config.vel_limit = vel_limit
        spec.config.duration = duration if isinstance(duration, float) else 2.0 / rate
        spec.config.checks = checks

        # Collision detector
        spec.config.collision = collision if isinstance(collision, dict) else None

        # Add converter & space_converter
        spec.inputs.goal.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", [-v for v in vel_limit], vel_limit, dtype="float32"
        )
        spec.inputs.position.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", "$(config lower)", "$(config upper)", dtype="float32"
        )
        spec.outputs.filtered.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", "$(config lower)", "$(config upper)", dtype="float32"
        )

    def initialize(
        self,
        joints: List[str],
        upper: List[float],
        lower: List[float],
        vel_limit: List[float],
        duration: float,
        checks: int,
        collision: dict,
    ):
        self.joints = joints
        self.upper = np.array(upper, dtype="float")
        self.lower = np.array(lower, dtype="float")
        self.vel_limit = np.array(vel_limit, dtype="float")
        self.duration = duration
        self.checks = checks
        self.dt = 1 / self.rate

        # Setup collision detector
        self.collision = collision
        if collision is not None:
            self.collision_check = True
            # Setup physics server for collision checking
            import pybullet

            if collision.get("gui", False):
                self.col_id = pybullet.connect(pybullet.GUI)
            else:
                self.col_id = pybullet.connect(pybullet.DIRECT)
            # Load workspace
            bodies = get_attribute_from_module(collision["workspace"])(self.col_id)
            # Generate robot urdf (if not a path but a text file)
            r = collision["robot"]
            if r["urdf"].endswith(".urdf"):  # Full path specified
                fileName = r["urdf"]
            else:  # First write to /tmp file (else pybullet cannot load urdf)
                import uuid  # Use this to generate a unique filename

                fileName = f"/tmp/{str(uuid.uuid4())}.urdf"
                with open(fileName, "w") as file:
                    file.write(r["urdf"])
            # Load robot
            bodies["robot"] = pybullet.loadURDF(
                fileName,
                basePosition=r.get("basePosition", None),
                baseOrientation=r.get("baseOrientation", None),
                useFixedBase=r.get("useFixedBase", True),
                flags=r.get("flags", 0),
                physicsClientId=self.col_id,
            )
            urdf = URDF.from_xml_file(fileName)
            # Determine collision pairs
            self_collision_pairs, workspace_pairs = col.get_named_collision_pairs(bodies, urdf, joints)
            # Create collision detector
            self.self_collision = col.CollisionDetector(self.col_id, bodies, joints, self_collision_pairs)
            self.workspace = col.CollisionDetector(self.col_id, bodies, joints, workspace_pairs)
            # Set distance at which objects are considered in collision.
            self.margin = collision.get("margin", 0.0)
            # self._test_collision_tester(joints)
        else:
            self.collision_check = False

    def _test_collision_tester(self, joints):
        while True:
            # compute shortest distances for a random configuration
            q = np.pi * (np.random.random(len(joints)) - 0.5)
            in_col = self.self_collision.in_collision(q)

            print(f"[self_collision] In collision = {in_col}")

            in_col = self.workspace.in_collision(q)

            print(f"[workspace] In collision = {in_col}")

            # wait for user to press enter to continue
            input("Press <enter>!")

    @register.states()
    def reset(self):
        self.safe_poses = deque(maxlen=10)
        self.consecutive_unsafe = 0

    @register.inputs(goal=Float32MultiArray, position=Float32MultiArray, velocity=Float32MultiArray)
    @register.outputs(filtered=Float32MultiArray, in_collision=UInt64)
    def callback(self, t_n: float, goal: Msg = None, position: Msg = None, velocity: Msg = None):
        goal = np.array(goal.msgs[-1].data, dtype="float32")
        position = np.array(position.msgs[-1].data, dtype="float32")
        velocity = np.array(velocity.msgs[-1].data, dtype="float32")

        # Setpoint last safe position
        if self.collision_check:
            if (
                self.self_collision.in_collision(q=position, margin=self.margin)
                and self.self_collision.get_distance().min() < 0
            ):
                in_collision = UInt64(data=1)
            elif self.workspace.in_collision(margin=self.margin) and self.workspace.get_distance().min() < 0:
                in_collision = UInt64(data=2)
            else:
                in_collision = UInt64(data=0)
                self.consecutive_unsafe = 0
                self.safe_poses.append(position)
        else:
            in_collision = UInt64(data=0)

        # Clip to velocity limits
        filtered = np.clip(goal, -self.vel_limit, self.vel_limit, dtype="float32")

        # Clip to joint limits
        max_position = position + filtered * self.dt
        clip_position = np.clip(max_position, self.lower, self.upper, dtype="float32")
        filtered = (clip_position - position) / self.dt

        if self.collision_check:
            # Linearly interpolate for intermediate joint configurations

            lin_vel = np.linspace(velocity, filtered, self.checks)
            interp = np.empty((position.shape[0], self.checks + 1), dtype="float32")

            interp[:, 0] = position
            interp[:, -1] = clip_position
            for i in range(self.checks):
                interp[:, i] += lin_vel[i] * self.dt / self.checks
                if i + 1 < self.checks:
                    interp[:, i + 1] = interp[:, i]

            for i in range(self.checks + 1):
                flag = self.self_collision.in_collision(q=interp[:, i], margin=self.margin)
                flag = flag | self.workspace.in_collision(margin=self.margin)
                if flag:
                    self.consecutive_unsafe += 1
                    if len(self.safe_poses) > 0:
                        idx = min(len(self.safe_poses), self.consecutive_unsafe + 1)
                        filtered = (self.safe_poses[-idx] - position) / self.dt
                        filtered = 0.5 * np.clip(filtered, -self.vel_limit, self.vel_limit, dtype="float32")
                    else:
                        filtered * 0
                    break
                elif i + 1 == self.checks:
                    self.consecutive_unsafe = 0
        else:
            self.consecutive_unsafe = 0
            in_collision = UInt64(data=0)
        return dict(filtered=Float32MultiArray(data=filtered), in_collision=in_collision)



# Adapted from github:@AdamHeins (02-04-2022)
# https://adamheins.com/blog/collision-detection-pybullet
# https://github.com/adamheins/pyb_utils/blob/4c610925acea69976cb5232169004af4c4de541a/pyb_utils/collision.py

from dataclasses import dataclass

import numpy as np
import pybullet as pyb


@dataclass
class NamedCollisionObject:
    """Name of a body and one of its links.
    The body name must correspond to the key in the `bodies` dict, but is
    otherwise arbitrary. The link name should match the URDF. The link name may
    also be None, in which case the base link (index -1) is used.
    """

    body_name: str
    link_name: str = None


@dataclass
class IndexedCollisionObject:
    """Index of a body and one of its links."""

    body_uid: int
    link_uid: int


@dataclass
class IndexedJointObject:
    """Index of a robot joint and its name."""

    joint_name: str
    joint_uid: int


def index_collision_pairs(physics_uid, bodies, named_collision_pairs):
    """Convert a list of named collision pairs to indexed collision pairs.
    In other words, convert named bodies and links to the indexes used by
    PyBullet to facilitate computing collisions between the objects.
    Parameters:
      physics_uid: Index of the PyBullet physics server to use.
      bodies: dict with body name keys and corresponding indices as values
      named_collision_pairs: a list of 2-tuples of NamedCollisionObject
    Returns: a list of 2-tuples of IndexedCollisionObject
    """

    # build a nested dictionary mapping body names to link names to link
    # indices
    body_link_map = {}
    for name, uid in bodies.items():
        body_link_map[name] = {}
        n = pyb.getNumJoints(uid, physics_uid)
        for i in range(n):
            info = pyb.getJointInfo(uid, i, physics_uid)
            link_name = info[12].decode("utf-8")
            body_link_map[name][link_name] = i

    def _index_named_collision_object(obj):
        """Map body and link names to corresponding indices."""
        body_uid = bodies[obj.body_name]
        if obj.link_name is not None:
            link_uid = body_link_map[obj.body_name][obj.link_name]
        else:  # todo: then make collision objects for each link.
            link_uid = -1
        return IndexedCollisionObject(body_uid, link_uid)

    # convert all pairs of named collision objects to indices
    indexed_collision_pairs = []
    for a, b in named_collision_pairs:
        a_indexed = _index_named_collision_object(a)
        b_indexed = _index_named_collision_object(b)
        indexed_collision_pairs.append((a_indexed, b_indexed))

    return indexed_collision_pairs


def index_joints(physics_uid, robot_id, joints):
    """Map a list of joint names to indexed joints.
    In other words, map named joints to the index used by
    PyBullet to facilitate setting the configuration.
    Parameters:
      physics_uid: Index of the PyBullet physics server to use.
      joints: list with joint name keys
    Returns: a list of IndexedJointObject
    """
    indexed_joints = []
    n = pyb.getNumJoints(robot_id, physics_uid)
    for joint_name in joints:
        for i in range(n):
            info = pyb.getJointInfo(robot_id, i, physics_uid)
            if joint_name == info[1].decode("utf-8"):
                indexed_joints.append(IndexedJointObject(joint_name, i))
                continue
    assert len(joints) == len(indexed_joints), "Not all joints were found in the provided urdf."
    return indexed_joints


class CollisionDetector:
    def __init__(self, col_id, bodies, joints, named_collision_pairs):
        self.ds = None
        self.col_id = col_id
        self.robot_id = bodies["robot"]
        self.indexed_collision_pairs = index_collision_pairs(self.col_id, bodies, named_collision_pairs)
        self.indexed_joints = index_joints(self.col_id, self.robot_id, joints)

    def compute_distances(self, q=None, max_distance=1.0):
        """Compute closest distances for a given configuration.
        Parameters:
          q: Iterable representing the desired configuration. This is applied
             directly to PyBullet body with index bodies["robot"].
          max_distance: Bodies farther apart than this distance are not queried
             by PyBullet, the return value for the distance between such bodies
             will be max_distance.
        Returns: A NumPy array of distances, one per pair of collision objects.
        """

        # put the robot in the given configuration
        if q is not None:
            for i, joint in enumerate(self.indexed_joints):
                pyb.resetJointState(self.robot_id, joint.joint_uid, q[i], physicsClientId=self.col_id)

        # compute shortest distances between all object pairs
        distances = []
        for a, b in self.indexed_collision_pairs:
            closest_points = pyb.getClosestPoints(
                a.body_uid,
                b.body_uid,
                distance=max_distance,
                linkIndexA=a.link_uid,
                linkIndexB=b.link_uid,
                physicsClientId=self.col_id,
            )

            # if bodies are above max_distance apart, nothing is returned, so
            # we just saturate at max_distance. Otherwise, take the minimum
            if len(closest_points) == 0:
                distances.append(max_distance)
            else:
                distances.append(np.min([pt[8] for pt in closest_points]))

        return np.array(distances)

    def in_collision(self, q=None, margin=0):
        """Returns True if configuration q is in collision, False otherwise.
        Parameters:
          q: Iterable representing the desired configuration.
          margin: Distance at which objects are considered in collision.
             Default is 0.0.
        """
        self.ds = self.compute_distances(q=q, max_distance=margin * 2)
        return (self.ds < margin).any()

    def get_distance(self):
        return self.ds


def get_robot_link_clusters(urdf, joints):
    clusters = {
        j: dict(links=[], connects=[]) for j in joints
    }  # {nf_joint: links=["l_1", "l_2"], connects=["joint_1", "join_2"]}
    world_cluster = "base"
    assert world_cluster not in joints, f"Cannot have a joint name '{world_cluster}'. It is a reserved joint name."
    clusters[world_cluster] = dict(links=[], connects=[])
    pm = urdf.parent_map
    cm = urdf.child_map

    def find_nonfixed_parent_joint(link):
        parent_joint, parent_link = pm[link]
        if parent_link not in pm:
            return world_cluster
        if parent_joint in joints:
            return parent_joint
        else:
            return find_nonfixed_parent_joint(parent_link)

    for link, (parent_joint, _parent_link) in urdf.parent_map.items():
        nf_parent_joint = find_nonfixed_parent_joint(link)
        if urdf.link_map[link].collision is not None:
            clusters[nf_parent_joint]["links"].append(link)
        clusters[nf_parent_joint]["connects"].append(parent_joint)
        if link in cm:
            for child_joint, _ in cm[link]:
                clusters[nf_parent_joint]["connects"].append(child_joint)

    for _, cluster in clusters.items():
        cluster["connects"] = list(set(cluster["connects"]))
    return clusters


def get_self_collision_pairs(clusters):
    key_pairs = {}
    for j_1, c_1 in clusters.items():
        key_pairs[j_1] = []
        for j_2, c_2 in clusters.items():
            if j_2 not in c_1["connects"] and j_1 not in c_2["connects"]:
                if j_2 not in key_pairs:
                    key_pairs[j_1].append(j_2)
    pairs = []
    for j_1, js in key_pairs.items():
        for l_1 in clusters[j_1]["links"]:
            nco_1 = NamedCollisionObject("robot", l_1)
            for j_2 in js:
                for l_2 in clusters[j_2]["links"]:
                    nco_2 = NamedCollisionObject("robot", l_2)
                    pairs.append((nco_1, nco_2))

    return pairs


def get_workspace_collision_pairs(bodies, clusters):
    pairs = []
    named_bodies = [NamedCollisionObject(b) for b in bodies if b != "robot"]
    for joint, cluster in clusters.items():
        for link in cluster["links"]:
            nco_1 = NamedCollisionObject("robot", link)
            for nco_2 in named_bodies:
                if joint == "base" and nco_2.body_name == "ground":
                    continue
                pairs.append((nco_1, nco_2))
    return pairs


def get_named_collision_pairs(bodies, urdf, joints):
    clusters = get_robot_link_clusters(urdf, joints)
    # Get self collision pairs
    self_collision_pairs = get_self_collision_pairs(clusters)
    # Get workspace pairs
    workspace_pairs = get_workspace_collision_pairs(bodies, clusters)
    return self_collision_pairs, workspace_pairs



from typing import Optional, List

# IMPORT ROS
from std_msgs.msg import UInt64, Float32MultiArray

# IMPORT EAGERX
from eagerx.core.constants import process
from eagerx.utils.utils import Msg
from eagerx.core.entities import EngineNode
import eagerx.core.register as register

# IMPORT INTERBOTIX
from interbotix_xs_modules import arm, gripper, core


def get_joint_indices(info, joints):
    joint_indices = []
    for n in joints:
        i = info.joint_names.index(n)
        joint_indices.append(info.joint_state_indices[i])
    return joint_indices


class XseriesSensor(EngineNode):
    @staticmethod
    @register.spec("XseriesSensor", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        joints: List[str],
        process: Optional[int] = process.NEW_PROCESS,
        color: Optional[str] = "cyan",
        mode: str = "position",
    ):
        """XseriesSensor spec"""
        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.inputs = ["tick"]
        spec.config.outputs = ["obs"]

        # Set parameters, defined by the signature of cls.initialize(...)
        spec.config.mode = mode
        spec.config.joints = joints

    def initialize(self, joints, mode):
        if not mode == "position":
            # todo: implement velocity mode
            raise NotImplementedError(f"This mode is not implemented: {mode}")
        self.joints = joints
        self.mode = mode
        self.dxl = core.InterbotixRobotXSCore(self.config["robot_type"], self.config["name"], False)
        self.arm = arm.InterbotixArmXSInterface(self.dxl, self.config["robot_type"], "arm", moving_time=0.2, accel_time=0.3)

        # Determine joint order
        self.joint_indices = get_joint_indices(self.dxl.robot_get_robot_info("group", "arm"), joints)

    @register.states()
    def reset(self):
        pass

    @register.inputs(tick=UInt64)
    @register.outputs(obs=Float32MultiArray)
    def callback(self, t_n: float, tick: Optional[Msg] = None):
        joint_state = self.dxl.robot_get_joint_states()
        obs = [joint_state.position[i] for i in self.joint_indices]
        return dict(obs=Float32MultiArray(data=obs))

    def close(self):
        pass


class XseriesArm(EngineNode):
    @staticmethod
    @register.spec("XseriesArm", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        joints: List[str],
        process: Optional[int] = process.NEW_PROCESS,
        color: Optional[str] = "green",
        mode: str = "position_control",
    ):
        """XseriesArm spec"""
        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.inputs = ["tick", "action"]
        spec.config.outputs = ["action_applied"]

        # Set parameters, defined by the signature of cls.initialize(...)
        spec.config.mode = mode
        spec.config.joints = joints

    def initialize(self, joints, mode):
        if not mode == "position_control":
            raise NotImplementedError(f"This mode is not implemented: {mode}")
        self.mode = mode
        self.dxl = core.InterbotixRobotXSCore(self.config["robot_type"], self.config["name"], False)
        self.arm = arm.InterbotixArmXSInterface(
            self.dxl, self.config["robot_type"], "arm", moving_time=4 / self.rate, accel_time=1 / self.rate
        )

        # Determine joint order
        self.joint_indices = get_joint_indices(self.dxl.robot_get_robot_info("group", "arm"), joints)

        self.arm.go_to_home_pose(moving_time=2.0, accel_time=0.3)

    @register.states()
    def reset(self):
        pass

    @register.inputs(tick=UInt64, action=Float32MultiArray)
    @register.outputs(action_applied=Float32MultiArray)
    def callback(
        self,
        t_n: float,
        tick: Optional[Msg] = None,
        action: Optional[Msg] = None,
    ):
        positions = action.msgs[-1].data
        indexed_positions = [positions[i] for i in self.joint_indices]
        self.arm.set_joint_positions(indexed_positions, moving_time=4 / self.rate, accel_time=1 / self.rate, blocking=False)
        # Send action that has been applied.
        return dict(action_applied=action.msgs[-1])

    def shutdown(self):
        self.dxl.robot_torque_enable("group", "arm", True)


class XseriesGripper(EngineNode):
    @staticmethod
    @register.spec("XseriesGripper", EngineNode)
    def spec(
        spec,
        name: str,
        rate: float,
        process: Optional[int] = process.NEW_PROCESS,
        color: Optional[str] = "green",
    ):
        """XseriesGripper spec"""
        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.inputs = ["tick", "action"]
        spec.config.outputs = ["action_applied"]

    def initialize(self):
        self.dxl = core.InterbotixRobotXSCore(self.config["robot_type"], self.config["name"], False)
        self.gripper = gripper.InterbotixGripperXSInterface(
            self.dxl,
            "gripper",
            gripper_pressure=0.5,
            gripper_pressure_lower_limit=150,
            gripper_pressure_upper_limit=350,
        )

    @register.states()
    def reset(self):
        pass

    @register.inputs(tick=UInt64, action=Float32MultiArray)
    @register.outputs(action_applied=Float32MultiArray)
    def callback(
        self,
        t_n: float,
        tick: Optional[Msg] = None,
        action: Optional[Msg] = None,
    ):
        # Maps action=[0, 1.] to [-gripper_value, gripper_value].
        gripper_value = self.gripper.gripper_value * ((action.msgs[-1].data[0] * 2) - 1)
        # Set gripper value
        self.gripper.gripper_controller(gripper_value, delay=0.0)
        # Send action that has been applied.
        return dict(action_applied=action.msgs[-1])

    def shutdown(self):
        self.gripper.open(delay=0.0)



from eagerx.core.entities import EngineState
import eagerx.core.register as register


class DummyState(EngineState):
    @staticmethod
    @register.spec("DummyState", EngineState)
    def spec(spec):
        pass

    def initialize(self):
        pass

    def reset(self, state, done):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eagerx_interbotix',
 'eagerx_interbotix.camera',
 'eagerx_interbotix.camera.pybullet',
 'eagerx_interbotix.camera.real',
 'eagerx_interbotix.reset',
 'eagerx_interbotix.safety',
 'eagerx_interbotix.solid',
 'eagerx_interbotix.solid.pybullet',
 'eagerx_interbotix.solid.real',
 'eagerx_interbotix.xseries',
 'eagerx_interbotix.xseries.pybullet',
 'eagerx_interbotix.xseries.real']

package_data = \
{'': ['*'],
 'eagerx_interbotix.camera': ['assets/*'],
 'eagerx_interbotix.solid': ['assets/*']}

install_requires = \
['eagerx-gui>=0.2.5,<0.3.0',
 'eagerx-pybullet>=0.1.8,<0.2.0',
 'eagerx-reality>=0.1.7,<0.2.0',
 'modern_robotics>=1.1.0,<2.0.0',
 'scipy>=1.0,<2.0',
 'stable-baselines3[extra]>=1.5.0,<2.0.0',
 'urdf-parser-py>=0.0.4,<0.0.5',
 'xacro>=1.13.3,<2.0.0']

setup_kwargs = {
    'name': 'eagerx-interbotix',
    'version': '0.1.6',
    'description': 'EAGERx interface to interbotix robot arms.',
    'long_description': None,
    'author': 'Jelle Luijkx',
    'author_email': 'j.d.luijkx@tudelft.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/eager-dev/eagerx_interbotix',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# ROS IMPORTS
from std_msgs.msg import Float32MultiArray, Float32

# EAGERx IMPORTS
from eagerx_pybullet.engine import PybulletEngine
from eagerx import Object, EngineNode, SpaceConverter, EngineState
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register


class Solid(Object):
    entity_id = "Solid"

    @staticmethod
    @register.sensors(
        pos=Float32MultiArray, vel=Float32MultiArray, orientation=Float32MultiArray, angular_vel=Float32MultiArray
    )
    @register.engine_states(
        pos=Float32MultiArray,
        vel=Float32MultiArray,
        orientation=Float32MultiArray,
        angular_vel=Float32MultiArray,
        lateral_friction=Float32,
    )
    @register.config(urdf=None, fixed_base=True, self_collision=True, base_pos=[0, 0, 0], base_or=[0, 0, 0, 1])
    def agnostic(spec: ObjectSpec, rate):
        """Agnostic definition of the Solid"""
        # Register standard converters, space_converters, and processors
        import eagerx.converters  # noqa # pylint: disable=unused-import

        # Position
        spec.sensors.pos.rate = rate
        spec.sensors.pos.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-999, -999, -999],
            high=[999, 999, 999],
        )
        spec.states.pos.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-1, -1, 0],
            high=[1, 1, 0],
        )

        # Velocity
        spec.sensors.vel.rate = rate
        spec.sensors.vel.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-10, -10, -10],
            high=[10, 10, 10],
        )
        spec.states.vel.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[0, 0, 0],
            high=[0, 0, 0],
        )

        # Orientation
        spec.sensors.orientation.rate = rate
        spec.sensors.orientation.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-1, -1, -1, -1],
            high=[1, 1, 1, 1],
        )
        spec.states.orientation.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[0, 0, -1, -1],
            high=[0, 0, 1, 1],
        )

        # Angular velocity
        spec.sensors.angular_vel.rate = rate
        spec.sensors.angular_vel.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-10, -10, -10],
            high=[10, 10, 10],
        )
        spec.states.angular_vel.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[0, 0, 0],
            high=[0, 0, 0],
        )

        # Dynamics
        spec.states.lateral_friction.space_converter = SpaceConverter.make(
            "Space_Float32",
            dtype="float32",
            low=0.1,
            high=0.5,
        )

    @staticmethod
    @register.spec(entity_id, Object)
    def spec(
        spec: ObjectSpec,
        name: str,
        urdf: str,
        sensors=None,
        states=None,
        rate=30,
        base_pos=None,
        base_or=None,
        self_collision=True,
        fixed_base=True,
    ):
        """Object spec of Solid"""
        # Modify default agnostic params
        # Only allow changes to the agnostic params (rates, windows, (space)converters, etc...
        spec.config.name = name
        spec.config.sensors = sensors if sensors is not None else ["pos", "vel", "orientation", "angular_vel"]
        spec.config.states = states if states is not None else ["pos", "vel", "orientation", "angular_vel"]

        # Add registered agnostic params
        spec.config.urdf = urdf
        spec.config.base_pos = base_pos if base_pos else [0, 0, 0]
        spec.config.base_or = base_or if base_or else [0, 0, 0, 1]
        spec.config.self_collision = self_collision
        spec.config.fixed_base = fixed_base

        # Add agnostic implementation
        Solid.agnostic(spec, rate)

    @staticmethod
    @register.engine(entity_id, PybulletEngine)
    def pybullet_engine(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (Pybullet) of the object."""
        # Import any object specific entities for this engine
        import eagerx_interbotix.solid.pybullet  # noqa # pylint: disable=unused-import
        import eagerx_pybullet  # noqa # pylint: disable=unused-import

        # Set object arguments (as registered per register.engine_params(..) above the engine.add_object(...) method.
        spec.PybulletEngine.urdf = spec.config.urdf
        spec.PybulletEngine.basePosition = spec.config.base_pos
        spec.PybulletEngine.baseOrientation = spec.config.base_or
        spec.PybulletEngine.fixed_base = spec.config.fixed_base
        spec.PybulletEngine.self_collision = spec.config.self_collision

        # Create engine_states (no agnostic states defined in this case)
        spec.PybulletEngine.states.pos = EngineState.make("LinkState", mode="position")
        spec.PybulletEngine.states.vel = EngineState.make("LinkState", mode="velocity")
        spec.PybulletEngine.states.orientation = EngineState.make("LinkState", mode="orientation")
        spec.PybulletEngine.states.angular_vel = EngineState.make("LinkState", mode="angular_vel")
        spec.PybulletEngine.states.lateral_friction = EngineState.make("PbDynamics", parameter="lateralFriction")

        # Create sensor engine nodes
        # Rate=None, but we will connect them to sensors (thus will use the rate set in the agnostic specification)
        pos = EngineNode.make("LinkSensor", "pos", rate=spec.sensors.pos.rate, process=2, mode="position")
        vel = EngineNode.make("LinkSensor", "vel", rate=spec.sensors.vel.rate, process=2, mode="velocity")
        orientation = EngineNode.make(
            "LinkSensor", "orientation", rate=spec.sensors.orientation.rate, process=2, mode="orientation"
        )
        angular_vel = EngineNode.make(
            "LinkSensor", "angular_vel", rate=spec.sensors.angular_vel.rate, process=2, mode="angular_vel"
        )

        # Create actuator engine nodes
        # Rate=None, but we will connect it to an actuator (thus will use the rate set in the agnostic specification)
        # Connect all engine nodes
        graph.add([pos, vel, orientation, angular_vel])
        graph.connect(source=pos.outputs.obs, sensor="pos")
        graph.connect(source=vel.outputs.obs, sensor="vel")
        graph.connect(source=orientation.outputs.obs, sensor="orientation")
        graph.connect(source=angular_vel.outputs.obs, sensor="angular_vel")

        # Check graph validity (commented out)
        # graph.is_valid(plot=True)



# ROS IMPORTS
from urdf_parser_py.urdf import URDF
from std_msgs.msg import Float32MultiArray

# EAGERx IMPORTS
from eagerx_reality.engine import RealEngine
from eagerx_pybullet.engine import PybulletEngine
from eagerx import Object, EngineNode, SpaceConverter, EngineState, Processor
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register
from eagerx_interbotix.utils import generate_urdf, get_configs


class Xseries(Object):
    entity_id = "Xseries"

    @staticmethod
    @register.sensors(pos=Float32MultiArray, vel=Float32MultiArray, ee_pos=Float32MultiArray)
    @register.actuators(pos_control=Float32MultiArray, vel_control=Float32MultiArray, gripper_control=Float32MultiArray)
    @register.engine_states(pos=Float32MultiArray, vel=Float32MultiArray, gripper=Float32MultiArray)
    @register.config(
        robot_type=None,
        joint_names=None,
        gripper_names=None,
        gripper_link=None,
        fixed_base=True,
        self_collision=True,
        base_pos=None,
        base_or=None,
        mode_config=None,
        motor_config=None,
        joint_lower=None,
        joint_upper=None,
        vel_limit=None,
        sleep_positions=None,
        urdf=None,
    )
    def agnostic(spec: ObjectSpec, rate):
        """Agnostic definition of the Xseries"""
        # Register standard converters, space_converters, and processors
        import eagerx.converters  # noqa # pylint: disable=unused-import

        # Set observation properties: (space_converters, rate, etc...)
        spec.sensors.pos.rate = rate
        spec.sensors.pos.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low="$(config joint_lower)",
            high="$(config joint_upper)",
        )
        spec.sensors.vel.rate = rate
        spec.sensors.vel.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-v for v in spec.config.vel_limit],
            high=spec.config.vel_limit,
        )
        spec.sensors.ee_pos.rate = rate
        spec.sensors.ee_pos.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-2, -2, 0],
            high=[2, 2, 2],
        )

        # Set actuator properties: (space_converters, rate, etc...)
        spec.actuators.pos_control.rate = rate
        spec.actuators.vel_control.rate = rate
        spec.actuators.gripper_control.rate = 1
        spec.actuators.vel_control.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-v for v in spec.config.vel_limit],
            high=spec.config.vel_limit,
        )
        spec.actuators.pos_control.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low="$(config joint_lower)",
            high="$(config joint_upper)",
        )
        spec.actuators.gripper_control.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", dtype="float32", low=[0], high=[1.0]
        )

        # Set model_state properties: (space_converters)
        spec.states.pos.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=len(spec.config.joint_names) * [0],
            high=len(spec.config.joint_names) * [0],
        )
        spec.states.vel.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=len(spec.config.joint_names) * [0],
            high=len(spec.config.joint_names) * [0],
        )
        spec.states.gripper.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray", dtype="float32", low=[0.5], high=[0.5]
        )

    @staticmethod
    @register.spec(entity_id, Object)
    def spec(
        spec: ObjectSpec,
        name: str,
        robot_type: str,
        sensors=None,
        actuators=None,
        states=None,
        rate=30,
        base_pos=None,
        base_or=None,
        self_collision=False,
        fixed_base=True,
        motor_config=None,
        mode_config=None,
    ):
        """Object spec of Xseries"""
        # Extract info on xseries arm from assets
        motor_config, mode_config = get_configs(robot_type, motor_config, mode_config)
        urdf = URDF.from_parameter_server(generate_urdf(robot_type, ns="pybullet_urdf"))

        # Sort joint_names according to joint_order
        joint_names, sleep_positions = [], []
        for n, sp in zip(motor_config["joint_order"], motor_config["sleep_positions"]):
            if n in motor_config["groups"]["arm"]:
                joint_names.append(n)
                sleep_positions.append(sp)
        motor_config["groups"]["arm"] = joint_names
        joint_names = motor_config["groups"]["arm"]
        gripper_names = [
            motor_config["grippers"]["gripper"]["left_finger"],
            motor_config["grippers"]["gripper"]["right_finger"],
        ]
        gripper_link = [link.name for link in urdf.links if "ee_gripper_link" in link.name][0]

        # Determine joint limits
        joint_lower, joint_upper, vel_limit = [], [], []
        for n in joint_names:
            joint_obj = next((joint for joint in urdf.joints if joint.name == n), None)
            joint_lower.append(joint_obj.limit.lower)
            joint_upper.append(joint_obj.limit.upper)
            vel_limit.append(joint_obj.limit.velocity)

        # Modify default config
        spec.config.name = name
        spec.config.sensors = sensors if sensors else ["pos"]
        spec.config.actuators = actuators if actuators else ["pos_control", "gripper_control"]
        spec.config.states = states if states else ["pos", "vel", "gripper"]

        # Add registered config params
        spec.config.robot_type = robot_type
        spec.config.joint_names = joint_names
        spec.config.gripper_names = gripper_names
        spec.config.gripper_link = gripper_link
        spec.config.base_pos = base_pos if base_pos else [0, 0, 0]
        spec.config.base_or = base_or if base_or else [0, 0, 0, 1]
        spec.config.self_collision = self_collision
        spec.config.fixed_base = fixed_base
        spec.config.motor_config = motor_config
        spec.config.mode_config = mode_config
        spec.config.joint_lower = joint_lower
        spec.config.joint_upper = joint_upper
        spec.config.vel_limit = vel_limit
        spec.config.sleep_positions = sleep_positions
        spec.config.urdf = urdf.to_xml_string()

        # Add agnostic implementation
        Xseries.agnostic(spec, rate)

    @staticmethod
    # This decorator pre-initializes engine implementation with default object_params
    @register.engine(entity_id, PybulletEngine)
    def pybullet_engine(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (Pybullet) of the object."""
        # Import any object specific entities for this engine
        import eagerx_interbotix.xseries.pybullet  # noqa # pylint: disable=unused-import
        import eagerx_pybullet  # noqa # pylint: disable=unused-import

        # Set object arguments (as registered per register.engine_params(..) above the engine.add_object(...) method.
        spec.PybulletEngine.urdf = generate_urdf(spec.config.robot_type, ns="pybullet_urdf")
        spec.PybulletEngine.basePosition = spec.config.base_pos
        spec.PybulletEngine.baseOrientation = spec.config.base_or
        spec.PybulletEngine.fixed_base = spec.config.fixed_base
        spec.PybulletEngine.self_collision = spec.config.self_collision

        # Determine gripper min/max
        urdf = URDF.from_parameter_server(spec.PybulletEngine.urdf)
        lower, upper = [], []
        for name in spec.config.gripper_names:
            joint_obj = next((joint for joint in urdf.joints if joint.name == name), None)
            lower.append(joint_obj.limit.lower)
            upper.append(joint_obj.limit.upper)
        constant = abs(lower[0])
        scale = upper[0] - lower[0]

        # Create engine_states (no agnostic states defined in this case)
        joints = spec.config.joint_names
        spec.PybulletEngine.states.gripper = EngineState.make("PbXseriesGripper", spec.config.gripper_names, constant, scale)
        spec.PybulletEngine.states.pos = EngineState.make("JointState", joints=joints, mode="position")
        spec.PybulletEngine.states.vel = EngineState.make("JointState", joints=joints, mode="velocity")

        # Fix gripper if we are not controlling it.
        if "gripper_control" not in spec.config.actuators:
            spec.PybulletEngine.states.gripper.fixed = True

        # Create sensor engine nodes
        # Rate=None, but we will connect them to sensors (thus will use the rate set in the agnostic specification)
        pos_sensor = EngineNode.make(
            "JointSensor", "pos_sensor", rate=spec.sensors.pos.rate, process=2, joints=joints, mode="position"
        )
        vel_sensor = EngineNode.make(
            "JointSensor", "vel_sensor", rate=spec.sensors.vel.rate, process=2, joints=joints, mode="velocity"
        )
        ee_pos_sensor = EngineNode.make(
            "LinkSensor",
            "ee_pos_sensor",
            rate=spec.sensors.ee_pos.rate,
            process=2,
            links=[spec.config.gripper_link],
            mode="position",
        )

        # Create actuator engine nodes
        # Rate=None, but we will connect it to an actuator (thus will use the rate set in the agnostic specification)
        pos_control = EngineNode.make(
            "JointController",
            "pos_control",
            rate=spec.actuators.pos_control.rate,
            process=2,
            joints=joints,
            mode="position_control",
            vel_target=len(joints) * [0.0],
            pos_gain=len(joints) * [0.5],
            vel_gain=len(joints) * [1.0],
            max_force=len(joints) * [2.0],
        )
        vel_control = EngineNode.make(
            "JointController",
            "vel_control",
            rate=spec.actuators.pos_control.rate,
            process=2,
            joints=joints,
            mode="velocity_control",
            vel_gain=len(joints) * [1.0],
            max_force=len(joints) * [2.0],
        )
        gripper = EngineNode.make(
            "JointController",
            "gripper_control",
            rate=spec.actuators.gripper_control.rate,
            process=2,
            joints=spec.config.gripper_names,
            mode="position_control",
            vel_target=[0.0, 0.0],
            pos_gain=[0.5, 0.5],
            vel_gain=[1.0, 1.0],
            max_force=[2.0, 2.0],
        )
        gripper.inputs.action.converter = Processor.make("MirrorAction", index=0, constant=constant, scale=scale)

        # Connect all engine nodes
        graph.add([pos_sensor, vel_sensor, ee_pos_sensor, pos_control, vel_control, gripper])
        graph.connect(source=pos_sensor.outputs.obs, sensor="pos")
        graph.connect(source=vel_sensor.outputs.obs, sensor="vel")
        graph.connect(source=ee_pos_sensor.outputs.obs, sensor="ee_pos")
        graph.connect(actuator="pos_control", target=pos_control.inputs.action)
        graph.connect(actuator="vel_control", target=vel_control.inputs.action)
        graph.connect(actuator="gripper_control", target=gripper.inputs.action)

        # Check graph validity (commented out)
        # graph.is_valid(plot=True)

    @staticmethod
    @register.engine(entity_id, RealEngine)
    def reality_engine(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (reality) of the object."""
        # Import any object specific entities for this engine
        import eagerx_interbotix.xseries.real  # noqa # pylint: disable=unused-import

        # Determine gripper min/max
        # Create engine_states (no agnostic states defined in this case)
        # todo: Where do we launch the driver file? In this engine? How to launch in namespace (based on robot_name + ns)?
        # todo: test with rviz (give as a separate option?)
        spec.RealEngine.states.gripper = EngineState.make("DummyState")
        spec.RealEngine.states.pos = EngineState.make("DummyState")
        spec.RealEngine.states.vel = EngineState.make("DummyState")

        # Create sensor engine nodes
        # Rate=None, but we will connect them to sensors (thus will use the rate set in the agnostic specification)
        joints = spec.config.joint_names
        pos_sensor = EngineNode.make(
            "XseriesSensor", "pos_sensor", rate=spec.sensors.pos.rate, joints=joints, process=2, mode="position"
        )
        vel_sensor = EngineNode.make(
            "XseriesSensor", "vel_sensor", rate=spec.sensors.vel.rate, joints=joints, process=2, mode="velocity"
        )

        # Create actuator engine nodes
        # Rate=None, but we will connect it to an actuator (thus will use the rate set in the agnostic specification)
        pos_control = EngineNode.make(
            "XseriesArm",
            "pos_control",
            rate=spec.actuators.pos_control.rate,
            joints=joints,
            process=2,
            mode="position_control",
        )
        gripper = EngineNode.make("XseriesGripper", "gripper_control", rate=spec.actuators.gripper_control.rate, process=2)

        # Connect all engine nodes
        graph.add([pos_sensor, vel_sensor, pos_control, gripper])
        graph.connect(source=pos_sensor.outputs.obs, sensor="pos")
        graph.connect(source=vel_sensor.outputs.obs, sensor="vel")
        graph.connect(actuator="pos_control", target=pos_control.inputs.action)
        graph.connect(actuator="gripper_control", target=gripper.inputs.action)



# ROS IMPORTS
from std_msgs.msg import Float32MultiArray
from sensor_msgs.msg import Image

# EAGERx IMPORTS
from eagerx_pybullet.engine import PybulletEngine
from eagerx import Object, EngineNode, SpaceConverter, EngineState
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register


class Camera(Object):
    entity_id = "Camera"

    @staticmethod
    @register.sensors(rgb=Image, pos=Float32MultiArray, orientation=Float32MultiArray)
    @register.engine_states(pos=Float32MultiArray, orientation=Float32MultiArray)
    @register.config(
        urdf=None,
        fixed_base=True,
        self_collision=True,
        base_pos=[0, 0, 0],
        base_or=[0, 0, 0, 1],
        render_shape=[480, 640],
        optical_link=None,
        calibration_link=None,
    )
    def agnostic(spec: ObjectSpec, rate):
        """Agnostic definition of the Camera"""
        # Register standard converters, space_converters, and processors
        import eagerx.converters  # noqa # pylint: disable=unused-import

        # Position
        spec.sensors.pos.rate = rate
        spec.sensors.pos.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-999, -999, -999],
            high=[999, 999, 999],
        )
        spec.states.pos.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[0.83, 0.0181, 0.75],
            high=[0.83, 0.0181, 0.75],
        )

        # Orientation
        spec.sensors.orientation.rate = rate
        spec.sensors.orientation.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[-1, -1, -1, -1],
            high=[1, 1, 1, 1],
        )
        spec.states.orientation.space_converter = SpaceConverter.make(
            "Space_Float32MultiArray",
            dtype="float32",
            low=[0.377, -0.04, -0.92, 0.088],
            high=[0.377, -0.04, -0.92, 0.088],
        )

        # Rgb
        spec.sensors.rgb.rate = rate
        spec.sensors.rgb.space_converter = SpaceConverter.make(
            "Space_Image",
            dtype="float32",
            low=0,
            high=1,
            shape=spec.config.render_shape + [3],
        )

    @staticmethod
    @register.spec(entity_id, Object)
    def spec(
        spec: ObjectSpec,
        name: str,
        sensors=None,
        states=None,
        rate=30,
        base_pos=None,
        base_or=None,
        self_collision=True,
        fixed_base=True,
        render_shape=None,
        urdf: str = None,
        optical_link: str = None,
        calibration_link: str = None,
    ):
        """Object spec of Camera"""
        # Modify default agnostic params
        # Only allow changes to the agnostic params (rates, windows, (space)converters, etc...
        spec.config.name = name
        spec.config.sensors = sensors if isinstance(sensors, list) else ["rgb"]
        spec.config.states = states if isinstance(states, list) else ["pos", "orientation"]

        # Add registered agnostic params
        spec.config.urdf = urdf
        spec.config.base_pos = base_pos if isinstance(base_pos, list) else [0, 0, 0]
        spec.config.base_or = base_or if isinstance(base_or, list) else [0, 0, 0, 1]
        spec.config.self_collision = self_collision
        spec.config.fixed_base = fixed_base
        spec.config.render_shape = render_shape if isinstance(render_shape, list) else [200, 300]
        spec.config.optical_link = optical_link if isinstance(optical_link, str) else None
        spec.config.calibration_link = calibration_link if isinstance(calibration_link, str) else None

        # Add agnostic implementation
        Camera.agnostic(spec, rate)

    @staticmethod
    @register.engine(entity_id, PybulletEngine)
    def pybullet_engine(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation (Pybullet) of the object."""
        # Import any object specific entities for this engine
        import eagerx_interbotix.solid.pybullet  # noqa # pylint: disable=unused-import
        import eagerx_pybullet  # noqa # pylint: disable=unused-import

        # Set object arguments (as registered per register.engine_params(..) above the engine.add_object(...) method.
        import pybullet_data

        urdf = spec.config.urdf
        spec.PybulletEngine.urdf = (
            urdf if isinstance(urdf, str) else "%s/%s.urdf" % (pybullet_data.getDataPath(), "cube_small")
        )
        # Initial position of baselink when urdf is loaded. Overwritten by state during the reset.
        spec.PybulletEngine.basePosition = spec.config.base_pos
        # Initial orientation of baselink when urdf is loaded. Overwritten by state during the reset.
        spec.PybulletEngine.baseOrientation = spec.config.base_or
        spec.PybulletEngine.fixed_base = spec.config.fixed_base
        spec.PybulletEngine.self_collision = spec.config.self_collision

        # Create engine_states (no agnostic states defined in this case)
        spec.PybulletEngine.states.pos = EngineState.make("LinkState", mode="position", link=spec.config.calibration_link)
        spec.PybulletEngine.states.orientation = EngineState.make(
            "LinkState", mode="orientation", link=spec.config.calibration_link
        )

        # Create sensor engine nodes
        # Rate=None, but we will connect them to sensors (thus will use the rate set in the agnostic specification)
        pos = EngineNode.make(
            "LinkSensor", "pos", rate=spec.sensors.pos.rate, process=2, mode="position", links=[spec.config.optical_link]
        )
        orientation = EngineNode.make(
            "LinkSensor",
            "orientation",
            rate=spec.sensors.orientation.rate,
            process=2,
            mode="orientation",
            links=[spec.config.optical_link],
        )
        rgb = EngineNode.make(
            "CameraSensor", "rgb", rate=spec.sensors.rgb.rate, process=2, mode="rgb", render_shape=spec.config.render_shape
        )
        rgb.config.inputs.append("pos")
        rgb.config.inputs.append("orientation")

        # Create actuator engine nodes
        # Rate=None, but we will connect it to an actuator (thus will use the rate set in the agnostic specification)
        # Connect all engine nodes
        graph.add([pos, orientation, rgb])
        graph.connect(source=pos.outputs.obs, sensor="pos")
        graph.connect(source=orientation.outputs.obs, sensor="orientation")
        graph.connect(source=rgb.outputs.image, sensor="rgb")
        graph.connect(source=pos.outputs.obs, target=rgb.inputs.pos)
        graph.connect(source=orientation.outputs.obs, target=rgb.inputs.orientation)

        # Check graph validity (commented out)
        # graph.is_valid(plot=True)



from eagerx.core.entities import EngineState
import eagerx.core.register as register
import pybullet


class PbXseriesGripper(EngineState):
    @staticmethod
    @register.spec("PbXseriesGripper", EngineState)
    def spec(spec, joints, constant, scale, fixed=False):
        spec.config.joints = joints
        spec.config.constant = constant
        spec.config.scale = scale
        spec.config.fixed = fixed

    def initialize(self, joints, constant, scale, fixed):
        self.obj_name = self.config["name"]
        flag = self.obj_name in self.simulator["robots"]
        assert flag, f'Simulator object "{self.simulator}" is not compatible with this simulation state.'
        self.joints = joints
        self.constant = constant
        self.scale = scale
        self.fixed = fixed
        self.robot = self.simulator["robots"][self.obj_name]
        self._p = self.simulator["client"]
        self.physics_client_id = self._p._client

        self.bodyUniqueId = []
        self.jointIndices = []
        for _idx, pb_name in enumerate(joints):
            bodyid, jointindex = self.robot.jdict[pb_name].get_bodyid_jointindex()
            self.bodyUniqueId.append(bodyid), self.jointIndices.append(jointindex)
        self.gripper_cb = self._gripper_reset(self._p, self.bodyUniqueId[0], self.jointIndices, constant, scale, fixed)

    def reset(self, state, done):
        if not done:
            self.gripper_cb(state.data)

    @staticmethod
    def _gripper_reset(p, bodyUniqueId, jointIndices, constant, scale, fixed):
        def cb(state):
            # Mirror & scale gripper position
            pos = scale * state[0] + constant
            gripper_pos = [pos, -pos]

            # Only 1-dof joints are supported here.
            # https://github.com/bulletphysics/bullet3/issues/2803
            velocities = []
            states = p.getJointStates(bodyUniqueId=bodyUniqueId, jointIndices=jointIndices, physicsClientId=p._client)
            for _i, (_, vel, _, _) in enumerate(states):
                velocities.append([vel])
            p.resetJointStatesMultiDof(
                targetValues=[[pos] for pos in gripper_pos],
                targetVelocities=velocities,
                bodyUniqueId=bodyUniqueId,
                jointIndices=jointIndices,
                physicsClientId=p._client,
            )
            # If we are not performing control with the gripper, fix the position.
            if fixed:
                p.setJointMotorControlArray(
                    bodyUniqueId=bodyUniqueId,
                    jointIndices=jointIndices,
                    controlMode=pybullet.VELOCITY_CONTROL,
                    forces=len(jointIndices) * [10**9],
                    physicsClientId=p._client,
                )

        return cb



__version__ = "0.1.6"
import eagerx_interbotix.reset  # noqa # pylint: disable=unused-import
import eagerx_interbotix.safety  # noqa # pylint: disable=unused-import
import eagerx_interbotix.xseries  # noqa # pylint: disable=unused-import
import eagerx_interbotix.solid  # noqa # pylint: disable=unused-import
import eagerx_interbotix.camera  # noqa # pylint: disable=unused-import
